# Autogenerated with SMOP 0.29
import numpy as np
from typing import Union
from geomaglib import legendre
#

def design_SHA_Sq_e_Re_v2(rho: Union[float, np.ndarray], theta: Union[float, np.ndarray], phi: Union[float, np.ndarray], nmax: int, mmax:int) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    # [A_r, A_theta, A_phi] = design_SHA_Sq_e_Re_v2(rho,theta,phi,t,t_ut,...
    #    nmax,mmax,p_vec,s_vec);

    # Calculate design matrices A_i that connects the vector of REAL,
    # Schmidt-normalized, spherical harmonic expansion coefficients,
    #        x = q_{nsp}^{m} and s_{nsp}^{m}
    # and the magnetic component B_r, B_theta, B_phi:
    #        B_r     = x'*A_r
    #        B_theta = x'*A_theta
    #        B_phi   = x'*A_phi
    # for the region below (primary) Sq currents.

    # Inputs:   rho(:)                  radius [units of reference radius]
    #           theta(:), phi(:)        co-latitude, longitude [deg]
    #           nmax, mmax              maximum degree and order

    # (Optimized version)

    # A. Chulliat, 2016-09-22
    # (from an earlier version dated 2011-04-21, with inputs from N. Olsen)
    rad = np.pi / 180
    N_data = np.size(theta, 0)
    if (np.size(phi, 0) != N_data):
        raise ValueError(
            "design_SHA_Sq_e: theta and phi have different dimensions"
        )
    if np.isscalar(theta) is not True and theta.ndim > 1:
        dim2 = np.size(theta, 1)
    else:
        dim2 = 1
    if dim2 > 1:
        theta = theta.transpose()
        phi = phi.transpose()

    cos_theta = np.cos(theta * rad)
    sin_theta = np.sin(theta * rad)
    # convert to row vector if input parameter is scalar
    if np.isscalar(rho) is True:
        rho = rho * np.ones(N_data)

    # number of parameters
    N_nm = mmax * (mmax + 2) + (nmax - mmax) * (2 * mmax + 1)
    # calculate sub-matrices for s=0, p=0
    A_r_0 = np.zeros((N_nm, N_data))
    A_theta_0 = np.zeros((N_nm, N_data))
    A_phi_0 = np.zeros((N_nm, N_data))
    # Cycles through all the coefficients for degree n, order m model.
    k = 0


    Leg = legendre.Flattened_Chaos_Legendre1(nmax, theta)
    Pnm, dPnm = Leg[0], Leg[1]

    # Cycle through all of n and m
    for n in range(1, nmax+1):
        rn1 = rho**(n-1)
        # The maximum m is less than the maximum n in the default DIFI
        Mmax = min(n, mmax)
        for m in range(Mmax+1):
            index = legendre.get_index(n, m, nmax)
            if m == 0:
                # no h terms for g10, g20, g30 etc...
                A_r_0[k] = (-n)*rn1*Pnm[index]
                A_theta_0[k] = -rn1*dPnm[index]
                A_phi_0[k] = -rn1*0
                k += 1
            else:
                alpha = m*phi*rad
                cos_phi = np.cos(alpha)
                sin_phi = np.sin(alpha)
                # g terms, as in g11, g21, g22 etc...
                A_r_0[k] = (-n)*rn1*Pnm[index]*cos_phi
                A_theta_0[k] = -rn1*dPnm[index]*cos_phi
                A_phi_0[k] = -rn1*Pnm[index]/sin_theta*(-m*sin_phi)
                k += 1
                # h terms as in h11, h21, h22 etc...
                A_r_0[k] = (-n)*rn1*Pnm[index]*sin_phi
                A_theta_0[k] = -rn1*dPnm[index]*sin_phi
                A_phi_0[k] = -rn1*Pnm[index]/sin_theta*m*cos_phi
                k += 1

    return A_r_0, A_theta_0, A_phi_0




import numpy as np
from DIFI.getSQfield import getSQfield
import pytest

def parse_geomagnetic_data(filepath):
    # Define column labels
    keys = [
        "year", "month", "day", "hour", "min", "sec",
        "r", "theta", "phi", "h", "lat", "lon", "F10.7",
        "B_r", "B_theta", "B_phi", "X", "Y", "Z"
    ]

    data = []
    with open(filepath, 'r') as f:
        for line in f:
            # Split the line into values and convert to floats
            values = list(map(float, line.strip().split()))
            # Create a dictionary for each row
            entry = dict(zip(keys, values))
            data.append(entry)

        for i in range(len(data)):
            data[i]["theta"] = 90 - data[i]["theta"]  # Convert theta to cotangent

    return data

def test_geocentric_input_accuracy():
    data = parse_geomagnetic_data("tests/test_values_DIFI8_v1_20250928.txt")
    # print(data[0])
    max_diff = 0.0

    for i in range(0, len(data)):

        B = getSQfield(data[i]['theta'],  data[i]['lon'], data[i]['year'], data[i]['month'], data[i]['day'], hour=data[i]['hour'], minutes = data[i]['min'], r = data[i]['r'],f107_1 = data[i]['F10.7'], geoc = True, model_name='difi8')
        if B['X'] != data[i]['X']:
            max_diff = max(max_diff, np.abs(B['X'] - data[i]['X']))
        if B['Y'] != data[i]['Y']:
            max_diff = max(max_diff, np.abs(B['Y'] - data[i]['Y']))
        #     print("Y mismatch", B['Y'] , data[i]['Y'], B['Y'] - data[i]['Y'])
        if B['Z'] != data[i]['Z']:
            max_diff = max(max_diff, np.abs(B['Z'] - data[i]['Z']))
        #     print("Z mismatch", B['Z'] , data[i]['Z'], B['Z'] - data[i]['Z'])

        # print(f"lon: {data[i]['lon']}, phi: {data[i]['phi']} lat: {data[i]['lat']}, theta: {data[i]['theta']} lat_np_degree: {np.rad2deg(90 - data[i]['theta'])}")
        if np.abs(B['X'] - data[i]['X']) > 1e-3:
            print("there is a mismatch in B['X'] - data[i]['X'] of size", B['X'] - data[i]['X'])
        if np.abs(B['Y'] - data[i]['Y']) > 1e-3:
            print("there is a mismatch in B['Y'] - data[i]['Y'] of size", B['Y'] - data[i]['Y'])
        if np.abs(B['Z'] - data[i]['Z']) > 1e-3:
            print("there is a mismatch in B['Z'] - data[i]['Z'] of size", B['Z'] - data[i]['Z'])

    print(f"Max difference: {max_diff}")
    assert(max_diff < 1e-3)


import warnings

import numpy as np
import os

from DIFI import jd2000_dt
from DIFI import forward_Sq_d_Re
# from DIFI import get_f107_index
from geomaglib import util, magmath
from typing import Optional, Union


def theta_to_geod_lat(theta: float) -> float:

    f = 1 / 298.257223563
    inv_f = 1 / (1 - f)
    lat = np.arctan(inv_f * inv_f * np.tan(np.radians(theta)))
    lat = np.degrees(lat)

    return lat

def getSQfield(lat: Union[float, list], lon: Union[float, list], year: Union[int, list], month: Union[int, list],
                day: Union[int, list], hour: Union[int, list]=0, minutes: Union[int, list]=0, 
                h: Union[float, list]=0,r: Union[float, list]=0, f107_1: Optional[Union[float, list]]=None, 
                model_name: Optional[Union[str]]="xdifi2", geoc:Optional[bool] = False, 
                return_geoc:Optional[bool] = False) -> dict:
    """
    Input:
        Latitude, lat (in WGS-84 coordinates)
        Longtitude, lon
        An array of year, year (Only good between 2000.0 and 2025.0
        An array of months, month
        An array of days, day

    Optional Input:
        An array of hours, hour
        An array of minutes, minutes
        Height above WGS84 ellipsoid, h
        Radius in geocentric coordinates, r
        Name of DIFI model, DIFI8,xDIFI2, DIFI7, model_name
        Treat inputs as geocentric (must use r input, not h), geoc
        Return geocentric B, return_geoc

    Output:
        B, the magnetic field due to the SQ in WGS-84 coordinates
    """

    earth_radius_km = 6371.2
    
    #forward_Sq_d_Re implicitly expects flat numpy arrays as inputs
    #some inputs are converted to equivalent numpy arrays by the
    #coordinate or time transformations, but there a few which 
    #need to be explicitly cast:
    lat = np.array(lat).flatten()
    lon = np.array(lon).flatten()
    h = np.array(h).flatten()
    
    if not geoc: #input `lat` is assumed geo*detic*, input `h` is used, input `r` is ignored
        r_gc, theta_gc = util.geod_to_geoc_lat(lat, h) #convert geodetic latitude and ellipsoidal height to geocentric lat and radius (distance from center of earth to your location)
    else: #input `lat` is assumed be geo*centric* latitude, input `r` is used as geocentric radius (distance from center of earth to your location), input `h` is ignored
        theta_gc = lat
        r_gc = r

    if np.any(np.abs((r_gc - 6371.2) - 110) < 30):
        warnings.warn("Warning: The altitude is within 30 km of 110 km, where the model switches between internal and external ionospheric current sources. The model is not expected to accurately represent the field within this transition zone.")

    cotheta_gc = 90 - theta_gc

    start_time = 0#5114.0

    sq_t = jd2000_dt.jd2000_dt(year, month, day, hour, minutes)
    warn_year_2001 = jd2000_dt.jd2000_dt(2001, 1, 1, 0, 0)
    warn_year_2014= jd2000_dt.jd2000_dt(2014, 1, 1, 0, 0)    
    warn_year_2024= jd2000_dt.jd2000_dt(2024, 1, 1, 0, 0)
    end_DIFI_valid = 9496.5

    if np.any(sq_t) > end_DIFI_valid:
        raise Exception(
            "DIFI is not valid after noon 12/31/2026 . Input time data contains a date outside DIFI's validity range."
        )
    start_f107_time = jd2000_dt.jd2000_dt(2000, 1, 1, 0, 0)
    end_f107_time =  jd2000_dt.jd2000_dt(2026, 1, 1, 12, 30)

    if (model_name.lower() == 'xdifi2'):
        """("importing coeff from xdifi")"""
        # from DIFI import get_f107_index_xDIFI as get_f107_index
        from DIFI.get_f107_index_all import get_f107_index, load_coefs, load_swarm_xDIFI
        difi_t_f107, difi_f107 = load_coefs()
        swarm_data = load_swarm_xDIFI()
        end_time = float(difi_t_f107[-1])
        if np.any(sq_t < warn_year_2001):
            warnings.warn("Dataset contains date before 2001.0, outside xDIFI2's recommended range", UserWarning)
        if np.any(sq_t >= warn_year_2024):

            warnings.warn("Dataset contains date after 2024.0, outside xDIFI2's recommended range", UserWarning)
    elif (model_name.lower() == 'difi8'):
        """("importing coeff from difi8")"""
        from DIFI.get_f107_index_all import get_f107_index, load_coefs, load_swarm_DIFI8
        difi_t_f107, difi_f107 = load_coefs()
        swarm_data = load_swarm_DIFI8()
        end_time = float(difi_t_f107[-1])
        if np.any(sq_t < warn_year_2014):
            warnings.warn("Dataset contains date before 2014.0, outside DIFI8's recommended range")
        if np.any(sq_t >= warn_year_2024):
            warnings.warn("Dataset contains date after 2024.0, outside DIFI8's recommended range")

    elif (model_name.lower() == 'difi7'):
        raise ValueError("Difi 7 is not supported since the release of difi8")
        """("importing coeff from difi8")"""
        from DIFI.get_f107_index_all import get_f107_index, load_coefs, load_swarm_DIFI7
        difi_t_f107, difi_f107 = load_coefs()
        swarm_data = load_swarm_DIFI7()
        end_time = float(difi_t_f107[-1])
    else:
        raise ValueError("Input model_name = difi8, xdifi2, or difi7. Input model name didn't match any models.")
    
    if f107_1 is None:
        f107_1 = get_f107_index(sq_t, start_f107_time, end_f107_time, difi_f107, difi_t_f107)
    else:
        f107_1 = np.array(f107_1).flatten()
        
    B_XYZ = {}
    # print "Difi input", r_gc, theta_gc, RV['lon'], sq_t, f107_1
    
    # calculate radii in units of reference radius
    a = 6371.2
    rho_Sq = (a + swarm_data['h']) / a
    rho = np.array(r_gc / a)
    if (np.min(rho) < rho_Sq) and (np.max(rho) >= rho_Sq):#If min below SQ and max above SQ
        #Split dataset into above and below SQ
        above_SQ_mask = rho < rho_Sq
        below_SQ_mask = rho >= rho_Sq
        # print('rioiarsnt\n\n\n',r_gc[above_SQ_mask], 
        # cotheta_gc[above_SQ_mask], lon[above_SQ_mask], sq_t[above_SQ_mask], f107_1[above_SQ_mask], 
        # swarm_data)
        [above_output1, above_output2] = forward_Sq_d_Re.forward_Sq_d_Re(r_gc[above_SQ_mask], 
        cotheta_gc[above_SQ_mask], lon[above_SQ_mask], sq_t[above_SQ_mask], f107_1[above_SQ_mask], 
        swarm_data)
        [below_output1, below_output2] = forward_Sq_d_Re.forward_Sq_d_Re(r_gc[below_SQ_mask], 
        cotheta_gc[below_SQ_mask], lon[below_SQ_mask], sq_t[below_SQ_mask], f107_1[below_SQ_mask], 
        swarm_data)
        B_1,B_2 = np.zeros((3, len(r_gc))), np.zeros((3, len(r_gc)))
        B_1[:, above_SQ_mask] = above_output1
        B_1[:, below_SQ_mask] = below_output1

        B_2[:, above_SQ_mask] = above_output2
        B_2[:, below_SQ_mask] = below_output2
    else:
        [B_1, B_2] = forward_Sq_d_Re.forward_Sq_d_Re(
            r_gc,
            cotheta_gc,
            lon,
            sq_t,
            f107_1,
            swarm_data
        )
    B_C = B_1 + B_2
    B_XYZ['Z'] = -1 * B_C[0]
    B_XYZ['Y'] = B_C[2]
    B_XYZ['X'] = -1 * B_C[1]

    if not return_geoc:
        if geoc:
            lat = theta_to_geod_lat(lat)
        Bx, By, Bz = magmath.rotate_magvec(B_XYZ['X'], B_XYZ['Y'], B_XYZ['Z'], theta_gc, lat)

        B_XYZ["X"] = Bx
        B_XYZ["Y"] = By
        B_XYZ["Z"] = Bz
    else:
        print("B_XYZ values are returning in geocentric (NEC) coord")
    return B_XYZ


import random
import unittest
import warnings
import numpy as np
import pytest

from DIFI import getSQfield

class test_validate_inputs(unittest.TestCase):
    def test_xdifi2_time_range(self):
        """
        Test the time range of the xDIFI2 model for point and list inputs.
        - warning if t < 2001.0 or t >= 2024.0
        - no warning if 2001.0 <= t < 2024.0
        - raise the error if t >= the last date of f107.DBL
        """

        N = 50
        lat = 25
        lon = 100
        years = np.linspace(1999, 2027, N)
        months = np.linspace(1, 12, N)
        days = np.linspace(1, 31, N)
        hours = np.linspace(0, 23, N)
        minutes = np.linspace(0, 59, N)
        h = 100

        idx_2000 = 0
        idx_2001 = N - 1
        idx_2024 = N - 1
        idx_end = N - 1

        # Test for the point input
        for i in range(N):

            # When the user provide f107 by himself, the end year can be extended after 2026.0
            B = getSQfield(lat, lon, years[i], months[i], days[i], f107_1=129, h=h, hour=0, minutes=0,
                           model_name="xDIFI2")

            if years[i] < 2000.0:
                idx_2000 = i
                with pytest.raises(Exception,
                                   match="This package does not contain f10.7 data before 01/01/2000. Input time data contains a date corresponding to f10.7 data not contained in this package"):
                    B = getSQfield(lat, lon, years[i], months[i], days[i], h=h, hour=0, minutes=0, model_name="xDIFI2")

            elif years[i] >= 2000.0 and years[i] < 2001.0:
                if idx_2001 == N - 1:
                    idx_2001 = i
                with pytest.warns(UserWarning,
                                  match="Dataset contains date after 2001.0, outside xDIFI2's recommended range"):

                    B = getSQfield(lat, lon, years[i], months[i], days[i], h=h, hour=0, minutes=0, model_name="xDIFI2")
                    warnings.warn("Dataset contains date after 2001.0, outside xDIFI2's recommended range", UserWarning)



            elif years[i] >= 2024.0 and years[i] < 2025.9:
                if idx_2024 == N - 1:
                    idx_2024 = i
                with pytest.warns(UserWarning,
                                  match="Dataset contains date after 2024.0, outside xDIFI2's recommended range"):

                    B = getSQfield(lat, lon, years[i], months[i], days[i], h=h, hour=0, minutes=0, model_name="xDIFI2")
                    warnings.warn("Dataset contains date after 2024.0, outside xDIFI2's recommended range", UserWarning)

            elif years[i] >= 2026.0:
                if idx_end == N - 1:
                    idx_end = i
                with pytest.raises(Exception,
                                   match="This package does not contain f10.7 data after noon 12/31/2025. Input time data contains a date corresponding to f10.7 data not contained in this package"):
                    B = getSQfield(lat, lon, years[i], months[i], days[i], h=h, hour=0, minutes=0, model_name="xDIFI2")

        # Test for the list input

        # When the user provide f107 by himself, the end year can be extended after 2026.0
        f107_arr = [129.0] * N
        B = getSQfield(lat, lon, years, months, days, f107_1=f107_arr, h=h, hour=0, minutes=0,
                       model_name="xDIFI2")

        with pytest.raises(Exception,
                           match="This package does not contain f10.7 data before 01/01/2000. Input time data contains a date corresponding to f10.7 data not contained in this package"):
            if idx_2000 < N - 1:
                B = getSQfield(lat, lon, years[:idx_2000 + 1], months[:idx_2000 + 1], days[:idx_2000 + 1], h=h,
                               hour=hours[:idx_2000 + 1], minutes=minutes[:idx_2000 + 1], model_name="xDIFI2")

        with pytest.warns(UserWarning,
                          match="Dataset contains date before 2001.0, outside xDIFI2's recommended range"):
            if idx_2001 < N - 1:
                B = getSQfield(lat, lon, years[idx_2000 + 1:idx_2001 + 1], months[idx_2000 + 1:idx_2001 + 1],
                               days[idx_2000 + 1:idx_2001 + 1], h=h,
                               hour=hours[idx_2000 + 1:idx_2001 + 1], minutes=minutes[idx_2000 + 1:idx_2001 + 1],
                               model_name="xdifi2")
                warnings.warn("Dataset contains date before 2001.0, outside xDIFI2's recommended range", UserWarning)

        with pytest.warns(UserWarning, match="Dataset contains date after 2024.0, outside xDIFI2's recommended range"):
            if idx_2024 < N - 1:
                B = getSQfield(lat, lon, years[idx_2024:idx_end], months[idx_2024:idx_end], days[idx_2024:idx_end], h=h,
                               hour=hours[idx_2024:idx_end], minutes=minutes[idx_2024:idx_end], model_name="xDIFI2")

                warnings.warn("Dataset contains date after 2024.0, outside xDIFI2's reccomended range", UserWarning)
        with pytest.raises(Exception,
                           match="This package does not contain f10.7 data after noon 12/31/2025. Input time data contains a date corresponding to f10.7 data not contained in this package."):
            B = getSQfield(lat, lon, years[idx_2024:], months[idx_2024:], days[idx_2024:], h=h, hour=hours[idx_2024:],
                           minutes=minutes[idx_2024:], model_name="xDIFI2")

    def test_difi8_time_range(self):

        """
            Test the time range of the DIFI8 model for point and list inputs.

            - warning if t < 2014.0 or t >= 2024.0
            - no warning if 2014.0 <= t < 2024.0
            - raise the error if t >= the last date of f107.DBL
        """

        N = 50
        lat = 25
        lon = 100
        years = np.linspace(1999, 2027, N)
        months = np.linspace(1, 12, N)
        days = np.linspace(1, 31, N)
        hours = np.linspace(0, 23, N)
        minutes = np.linspace(0, 59, N)
        h = 100

        # Initialize indices for the first year with data, the year with a warning, and the last year
        idx_2000 = 0
        idx_2014 = N - 1
        idx_2024 = N - 1
        idx_end = N - 1

        # Test for the point input
        for i in range(N):

            # When the user provide f107 by himself, the end year can be extended after 2026.0
            B = getSQfield(lat, lon, years[i], months[i], days[i], f107_1=129, h=h, hour=0, minutes=0,
                           model_name="difi8")

            if years[i] < 2000.0:
                idx_2000 = i
                with pytest.raises(Exception,
                                   match="This package does not contain f10.7 data before 01/01/2000. Input time data contains a date corresponding to f10.7 data not contained in this package"):
                    B = getSQfield(lat, lon, years[i], months[i], days[i], h=h, hour=0, minutes=0, model_name="difi8")

            elif 2000.0 <= years[i] < 2014.0:

                if idx_2014 == N - 1:
                    idx_2014 = i
                with pytest.warns(UserWarning,
                                  match="Dataset contains date before 2014.0, outside DIFI8's reccomended range"):
                    B = getSQfield(lat, lon, years[i], months[i], days[i], h=h, hour=0, minutes=0, model_name="difi8")
                    warnings.warn("Dataset contains date before 2014.0, outside DIFI8's reccomended range", UserWarning)

            elif years[i] > 2024.0 and years[i] < 2025.9:
                if idx_2024 == N - 1:
                    idx_2024 = i

                with pytest.warns(UserWarning,
                                  match="Dataset contains date after 2024.0, outside DIFI8's recommended range"):

                    B = getSQfield(lat, lon, int(years[i]), int(months[i]), int(days[i]), h=h, hour=0, minutes=0,
                                   model_name="difi8")
                    warnings.warn("Dataset contains date before 2024.0, outside DIFI8's recommended range", UserWarning)


            elif years[i] >= 2026.0:
                if idx_end == N - 1:
                    idx_end = i

                with pytest.raises(Exception,
                                   match="This package does not contain f10.7 data after noon 12/31/2025. Input time data contains a date corresponding to f10.7 data not contained in this package"):
                    B = getSQfield(lat, lon, years[i], months[i], days[i], h=h, hour=0, minutes=0, model_name="difi8")

        # Test for the list input

        # When the user provide f107 by himself, the end year can be extended after 2026.0
        f107_arr = [129.0] * N
        B = getSQfield(lat, lon, years, months, days, f107_1=f107_arr, h=h, hour=0, minutes=0,
                       model_name="difi8")

        with pytest.raises(Exception,
                           match="This package does not contain f10.7 data before 01/01/2000. Input time data contains a date corresponding to f10.7 data not contained in this package"):

            if idx_2000 < N - 1:
                B = getSQfield(lat, lon, years[:idx_2000 + 1], months[:idx_2000 + 1], days[:idx_2000 + 1], h=h,
                               hour=hours[:idx_2014 + 1], minutes=minutes[:idx_2014 + 1], model_name="difi8")

        with pytest.warns(UserWarning,
                          match="Dataset contains date before 2014.0, outside DIFI8's recommended range"):

            if idx_2014 < N - 1:
                B = getSQfield(lat, lon, years[idx_2000 + 1:idx_2014 + 1], months[idx_2000 + 1:idx_2014 + 1],
                               days[idx_2000 + 1:idx_2014 + 1], h=h,
                               hour=hours[idx_2000 + 1:idx_2014 + 1], minutes=minutes[idx_2000 + 1:idx_2014 + 1],
                               model_name="difi8")
                warnings.warn("Dataset contains date before 2014.0, outside DIFI8's recommended range", UserWarning)

        with pytest.warns(UserWarning, match="Dataset contains date after 2024.0, outside DIFI8's recommended range"):
            if idx_2024 < N - 1:
                B = getSQfield(lat, lon, years[idx_2024:idx_end], months[idx_2024:idx_end], days[idx_2024:idx_end], h=h,
                               hour=hours[idx_2024:idx_end], minutes=minutes[idx_2024:idx_end], model_name="difi8")
                warnings.warn("Dataset contains date before 2024.0, outside DIFI8's recommended range", UserWarning)

        with pytest.raises(Exception,
                           match="This package does not contain f10.7 data after noon 12/31/2025. Input time data contains a date corresponding to f10.7 data not contained in this package."):
            B = getSQfield(lat, lon, years[idx_2024:], months[idx_2024:], days[idx_2024:], h=h, hour=hours[idx_2024:],
                           minutes=minutes[idx_2024:], model_name="difi8")

    def test_altitude_range(self):

        """
        Both DIFI8 and xDIFI2 should allow altitiude from -1 to 1000 km
        """

        N = 50
        lat = [45]
        lon = [100]
        years = np.linspace(2000., 2025, N)
        months = np.linspace(1., 12., N)
        days = np.linspace(1., 30., N)
        hours = np.linspace(0., 23., N)
        minutes = np.linspace(0., 59., N)
        #h = np.linspace(-1, 1000, N)
        h = 100


        B = getSQfield(lat, lon, years, months, days, h=h, hour=hours,
                           minutes=minutes, model_name="difi8")

        B = getSQfield(lat, lon, years, months, days, h=h, hour=hours,
                           minutes=minutes, model_name="xdifi2")

    def test_endtime(self):
        """
        The end_time should allow the last date of the year of 2025
        """

        N = 50
        lat = 25
        lon = 100
        h = 1000

        with pytest.warns(UserWarning, match="Dataset contains date after 2024.0, outside xDIFI2's reccomended range"):
            B = getSQfield(lat, lon, year=2024, month=1, day=1, h=h, hour=0,
                           minutes=0, model_name="xdifi2")
            warnings.warn("Dataset contains date after 2024.0, outside xDIFI2's reccomended range", UserWarning)


if __name__ == '__main__':
    unittest.main()



import numpy as np
from DIFI.getSQfield import getSQfield
import pytest


def parse_geomagnetic_data(filepath):
    # Define column labels
    keys = [
        "year", "month", "day", "hour", "min", "sec",
        "r", "theta", "phi", "h", "lat", "lon", "F10.7",
        "B_r", "B_theta", "B_phi", "X", "Y", "Z"
    ]

    data = []
    with open(filepath, 'r') as f:
        for line in f:
            # Split the line into values and convert to floats
            values = list(map(float, line.strip().split()))
            # Create a dictionary for each row
            entry = dict(zip(keys, values))
            data.append(entry)
    
    return data
def test_above_and_below_SQ():
    models = ['xDIFI2', 'DIFI8']
    for model_name in models:
        data = parse_geomagnetic_data(f"tests/test_values_{model_name}_v1_20250928.txt")
        lat, lon, year, month, day, hour, minute, h, f107 = [],[],[],[],[],[],[],[],[]
        for i in range(0,len(data)):
            lat.append(data[i]['lat'])
            lon.append(data[i]['lon'])
            year.append(data[i]['year'])
            month.append(data[i]['month'])
            day.append(data[i]['day'])
            hour.append(data[i]['hour'])
            minute.append(data[i]['min'])
            h.append(data[i]['h'])
            f107.append(data[i]['F10.7'])
        #================================================================================================
        #Create test values
        #================================================================================================
        #Create values below SQ field
        h = np.linspace(80, 105, len(h))

        B = getSQfield(lat, lon, year, 
                        month, day, hour=hour,
                            minutes = minute, h = h,f107_1 = f107, 
                            model_name = model_name)
        B = np.array([B['X'], B['Y'], B['Z']])
        np.save(f"tests/{model_name}_below_SQ_test_values", B)
        h = np.linspace(115, 140, len(h))
        B = getSQfield(lat, lon, year, 
                        month, day, hour=hour,
                            minutes = minute, h = h,f107_1 = f107, 
                            model_name = model_name)
        B = np.array([B['X'], B['Y'], B['Z']])
        np.save(f"tests/{model_name}_after_SQ_test_values", B)
        
        max_diff = 0.0
        
        temp_lat = []
        temp_lon = []
        temp_year = []
        temp_month = []
        temp_day = []
        temp_hour = []
        temp_minute = []
        temp_h = []
        temp_f107 = []
        temp_model_name = []

        temp_lat.extend(lat)
        temp_lon.extend(lon)
        temp_year.extend(year)
        temp_month.extend(month)
        temp_day.extend(day)
        temp_hour.extend(hour)
        temp_minute.extend(minute)
        temp_h.extend(np.linspace(80, 105, len(h)))
        temp_f107.extend(f107)

        temp_lat.extend(lat)
        temp_lon.extend(lon)
        temp_year.extend(year)
        temp_month.extend(month)
        temp_day.extend(day)
        temp_hour.extend(hour)
        temp_minute.extend(minute)
        temp_h.extend(np.linspace(115, 140, len(h)))
        temp_f107.extend(f107)
        lat = temp_lat
        lon = temp_lon
        year = temp_year
        month = temp_month
        day = temp_day
        hour = temp_hour
        minute = temp_minute
        h = temp_h
        f107 = temp_f107


        B = getSQfield(lat, lon, year, 
                        month, day, hour=hour,
                            minutes = minute, h = h,f107_1 = f107, 
                            model_name = model_name)
        below_ans = np.load(f"tests/{model_name}_below_SQ_test_values.npy", allow_pickle= False)
        above_ans = np.load(f"tests/{model_name}_after_SQ_test_values.npy", allow_pickle= False)
        ansX = []
        ansY = []
        ansZ = []
        ansX.extend(below_ans[0])
        ansY.extend(below_ans[1])
        ansZ.extend(below_ans[2])
        
        ansX.extend(above_ans[0])
        ansY.extend(above_ans[1])
        ansZ.extend(above_ans[2])
        for i in range(0, len(ansX)):
            max_diff = max(max_diff, np.abs(ansX[i] - B["X"][i]))
            max_diff = max(max_diff, np.abs(ansY[i] - B["Y"][i]))
            max_diff = max(max_diff, np.abs(ansZ[i] - B["Z"][i]))
        
            # if not i // 10:
            #     print(np.abs(ansX[i] - B["X"][i]), ansX[i] ,  B["X"][i])
            #     print(np.abs(ansY[i] - B["Y"][i]), ansY[i] ,  B["Y"][i])
            #     print(np.abs(ansZ[i] - B["Z"][i]), ansZ[i] ,  B["Z"][i])
        B = getSQfield(lat[0], lon[0], year[0], 
                        month[0], day[0], hour=hour[0],
                            minutes = minute[0], h = h[0],f107_1 = f107[0], 
                            model_name = model_name)
        below_ans = np.load(f"tests/{model_name}_below_SQ_test_values.npy", allow_pickle= False)
        above_ans = np.load(f"tests/{model_name}_after_SQ_test_values.npy", allow_pickle= False)
        ansX = []
        ansY = []
        ansZ = []
        ansX.extend(below_ans[0])
        ansY.extend(below_ans[1])
        ansZ.extend(below_ans[2])
        
        ansX.extend(above_ans[0])
        ansY.extend(above_ans[1])
        ansZ.extend(above_ans[2])
        #Create inputs which have span above and below the SQ field
        
        # for i in range(0, len(data)):
        #     # B = getSQfield(data[i]['lat'], data[i]['lon'], data[i]['year'], data[i]['month'], data[i]['day'], hour=data[i]['hour'], minutes = data[i]['min'], h= data[i]['h'],f107_1 = data[i]['F10.7'], model_name='xdifi2')

        #     if B['X'][i] != data[i]['X']:
        #         max_diff = max(max_diff, np.abs(B['X'][i] - data[i]['X']))
        #     if B['Y'][i] != data[i]['Y']:
        #         max_diff = max(max_diff, np.abs(B['Y'][i] - data[i]['Y']))
        #     #     print("Y mismatch", B['Y'][i] , data[i]['Y'], B['Y'][i] - data[i]['Y'])
        #     if B['Z'][i] != data[i]['Z']:
        #         max_diff = max(max_diff, np.abs(B['Z'][i] - data[i]['Z']))

        #     if np.abs(B['X'][i] - data[i]['X']) > 1e-3:
        #         print("there is a mismatch in B['X'][i] - data[i]['X'] of size", B['X'][i] - data[i]['X'])
        #         raise ValueError("There was a mismatch of magnitude greater than 1e-3 in the X component")
        #     if np.abs(B['Y'][i] - data[i]['Y']) > 1e-3:
        #         print("there is a mismatch in B['Y'][i] - data[i]['Y'] of size", B['Y'][i] - data[i]['Y'])
        #         raise ValueError("There was a mismatch of magnitude greater than 1e-3 in the Y component")
        #     if np.abs(B['Z'][i] - data[i]['Z']) > 1e-3:
        #         print("there is a mismatch in B['Z'][i] - data[i]['Z'][i] of size", B['Z'][i] - data[i]['Z'])
        #         raise ValueError("There was a mismatch of magnitude greater than 1e-3 in the Y component")
        print(f"Max difference of {model_name} when calculating above and below SQ field simultaneously vs individually input: {max_diff}. Should be zero")
        assert(max_diff < 1e-12)



import math
import numpy as np



def gg2gm_2010(
    theta_gg: np.ndarray=None,
    phi_gg: np.ndarray=None,
    theta_gm=None,
    phi_gm=None,
    B_theta=None,
    B_phi=None,
    get_R: bool=False, 
    theta_b = 9.92,
    phi_b = 287.78
):
    # transformation between geographic (geocentric as opposed to geodetic)
    # and geomagnetic (dipole) co-ordinates
    #          and components

    # theta_gm                = gg2gm(theta_gg, phi_gg)
    # [theta_gm, phi_gm]      = gg2gm(theta_gg, phi_gg)
    # [theta_out, phi_out]    = gg2gm(theta_in, phi_in, i_trans)
    #                         i_trans = +1: gg -> gm
    #                         i_trans = -1: gm -> gg

    # theta(:) co-latitude [deg]
    # phi(:)   longitude [deg]

    # R(:,2,2) is the matrix that rotates a horizontal vector from GEO to MAG:
    #          [B_theta_gm; B_phi_gm] = R*[B_theta_geo; B_phi_geo]

    # March 2003, Nils Olsen, DSRI

    # September 2004: Output of matrix R added

    # February 2011: coordinates of geomagnetic North pole updated to 2010.0
    # (Arnaud Chulliat, IPGP)
    # December 2016: translated to python
    # (Adam Woods, NCEI)
    # October 2025: changed pole location to be argument | Default IGRF 2010 pole
    # (Collin Kadlecek, NCEI)

    rad = math.pi / 180

    if (
        (theta_gg is not None and phi_gg is not None)
        and (theta_gm is None or phi_gm is None)
    ):
        i_trans = 1
        theta_in = theta_gg
        phi_in = phi_gg
    elif (
        (theta_gg is None and phi_gg is None)
        and (theta_gm is not None or phi_gm is not None)
    ):
        i_trans = -1
        theta_in = theta_gm
        phi_in = phi_gm
    else:
        print(
            "Need either values for phi_gg and theta_gg xor phi_gm and "
            "theta_gm, not neither or both"
        )

    s_p_b = np.sin(phi_b*rad)
    c_p_b = np.cos(phi_b*rad)
    c_t_b = np.cos(theta_b*rad)
    s_t_b = np.sin(theta_b*rad)
    A = np.array(
        [
            [c_t_b*c_p_b, c_t_b*s_p_b,  -s_t_b],
            [-s_p_b, c_p_b, 0],
            [s_t_b*c_p_b, s_t_b*s_p_b, c_t_b],
        ]
    )

    if i_trans == - 1:
        A = A.transpose()

    c_t = np.cos(theta_in*rad)
    s_t = np.sin(theta_in*rad)
    c_p = np.cos(phi_in*rad)
    s_p = np.sin(phi_in*rad)
    z = c_t
    x = s_t*c_p
    y = s_t*s_p
    x_out = A[0, 0] * x + A[0, 1] * y + A[0, 2] * z
    y_out = A[1, 0] * x + A[1, 1] * y + A[1, 2] * z
    z_out = A[2, 0] * x + A[2, 1] * y + A[2, 2] * z
    theta_out = 90 - np.arctan2(z_out, np.sqrt(x_out ** 2 + y_out ** 2)) / rad
    phi_out = (np.arctan2(y_out, x_out) / rad) % 360

    # Still need to pythonify the below...
    if B_theta is not None:
        raise NotImplementedError(
            "The rotation of the Magnetic field vector is not implemented"
        )
        BE = B_theta*c_t
        Bx = BE*c_p - B_phi*s_p
        By = BE*s_p + B_phi*c_p
        Bz = -B_theta*s_t
        Bx_gm = A[0, 0] * Bx + A[0, 1] * By + A[0, 2] * Bz
        By_gm = A[1, 0] * Bx + A[1, 1] * By + A[1, 2] * Bz
        Bz_gm = A[2, 0] * Bx + A[2, 1] * By + A[2, 2] * Bz
        c_t = np.cos(theta_gm*rad)
        s_t = np.sin(theta_gm*rad)
        c_p = np.cos(phi_gm*rad)
        s_p = np.sin(phi_gm*rad)
        BE = Bx_gm*c_p + By_gm*s_p
        varargout[1] = np.multiply(BE, c_t) - np.multiply(Bz_gm, s_t)
        varargout[2] = np.multiply(By_gm, c_p) - np.multiply(Bx_gm, s_p)

    # Still need to pythonify the above
    if get_R:
        if i_trans != 1:
            raise ValueError(r"Calculation of R only for i_trans = 1")
        theta_gm = theta_out
        phi_gm = phi_out
        c_t = np.cos(theta_gg*rad)
        s_t = np.sin(theta_gg*rad)
        c_p = np.cos(phi_gg*rad)
        s_p = np.sin(phi_gg*rad)
        R_1 = np.zeros((np.size(c_p, 0), 3, 3))
        R_1[:, 0, :] = np.array([s_t*c_p, c_t*c_p, -s_p]).transpose()
        R_1[:, 1, :] = np.array([s_t*s_p, c_t*s_p, c_p]).transpose()
        R_1[:, 2, :] = np.array(
            [c_t, -s_t, np.zeros(np.size(c_p))]
        ).transpose()
        c_t = np.cos(theta_gm*rad)
        s_t = np.sin(theta_gm*rad)
        c_p = np.cos(phi_gm*rad)
        s_p = np.sin(phi_gm*rad)
        R_2 = np.zeros((np.size(c_p, 0), 3, 3))
        R_2[:, 0, :] = np.array([s_t*c_p, s_t*s_p, c_t]).transpose()
        R_2[:, 1, :] = np.array([c_t*c_p, c_t*s_p, -s_t]).transpose()
        R_2[:, 2, :] = np.array(
            [-s_p, c_p, np.zeros(np.size(c_p))]
        ).transpose()
        s_p_b = s_p_b*np.ones(np.size(c_p, 0))
        c_p_b = c_p_b*np.ones(np.size(c_p, 0))
        s_t_b = s_t_b*np.ones(np.size(c_p, 0))
        c_t_b = c_t_b*np.ones(np.size(c_p, 0))
        # While they look like extra parentheticals these are necessary as
        # zeros takes a tuple as a parameter
        R_mag_geo = np.zeros((np.size(c_p, 0), 3, 3))
        R_mag_geo[:, 0, :] = np.array(
            [c_t_b*c_p_b, c_t_b*s_p_b, -s_t_b]
            ).transpose()
        R_mag_geo[:, 1, :] = np.array(
            [-s_p_b, c_p_b, np.zeros(np.size(c_p, 0))]
            ).transpose()
        R_mag_geo[:, 2, :] = np.array(
            [s_t_b*c_p_b, s_t_b*s_p_b, c_t_b]
            ).transpose()
        R_tmp = np.einsum('ijk,ikl->ijl', R_mag_geo, R_1)
        R_3 = np.einsum('ijk,ikl->ijl', R_2, R_tmp)
        sheet_1, sheet_2, sheet_3 = np.dsplit(R_3, 3)
        sm_R3 = np.dstack((sheet_2, sheet_3))
        column_1, column_2, column_3 = np.hsplit(sm_R3, 3)
        sm_R3 = np.hstack((column_2, column_3))
        # peeling off top sheet and column
        return [theta_out, phi_out, sm_R3]
    else:
        return [theta_out, phi_out]



import numpy as np
# from DIFI import get_SQ_field_TESTING_ONLY as getSQfield
from DIFI.getSQfield import getSQfield
import pytest

def parse_geomagnetic_data(filepath):
    # Define column labels
    keys = [
        "year", "month", "day", "hour", "min", "sec",
        "r", "theta", "phi", "h", "lat", "lon", "F10.7",
        "B_r", "B_theta", "B_phi", "X", "Y", "Z"
    ]

    data = []
    with open(filepath, 'r') as f:
        for line in f:
            # Split the line into values and convert to floats
            values = list(map(float, line.strip().split()))
            # Create a dictionary for each row
            entry = dict(zip(keys, values))
            data.append(entry)
    
    return data
def test_xdifi2_accuracy():
    data = parse_geomagnetic_data("tests/test_values_xDIFI2_v1_20250928.txt")
    max_diff = 0.0

    for i in range(0, len(data)):
        B = getSQfield(data[i]['lat'], data[i]['lon'], data[i]['year'], data[i]['month'], data[i]['day'], hour=data[i]['hour'], minutes = data[i]['min'], h= data[i]['h'],f107_1 = data[i]['F10.7'], model_name='xdifi2')

        if B['X'] != data[i]['X']:
            max_diff = max(max_diff, np.abs(B['X'] - data[i]['X']))
        if B['Y'] != data[i]['Y']:
            max_diff = max(max_diff, np.abs(B['Y'] - data[i]['Y']))
        #     print("Y mismatch", B['Y'] , data[i]['Y'], B['Y'] - data[i]['Y'])
        if B['Z'] != data[i]['Z']:
            max_diff = max(max_diff, np.abs(B['Z'] - data[i]['Z']))

        if np.abs(B['X'] - data[i]['X']) > 1e-3:
            print("there is a mismatch in B['X'] - data[i]['X'] of size", B['X'] - data[i]['X'])
            raise ValueError("There was a mismatch of magnitude greater than 1e-3 in the X component")
        if np.abs(B['Y'] - data[i]['Y']) > 1e-3:
            print("there is a mismatch in B['Y'] - data[i]['Y'] of size", B['Y'] - data[i]['Y'])
            raise ValueError("There was a mismatch of magnitude greater than 1e-3 in the Y component")
        if np.abs(B['Z'] - data[i]['Z']) > 1e-3:
            print("there is a mismatch in B['Z'] - data[i]['Z'] of size", B['Z'] - data[i]['Z'])
            raise ValueError("There was a mismatch of magnitude greater than 1e-3 in the Y component")
    print(f"Max difference: {max_diff}")
    assert(max_diff < 1e-3)



import datetime
import numpy as np
from typing import Union


def jd2000_dt(year: Union[float, list], month: Union[float, list], day: Union[float, list], UT: Union[float, list]=0, minutes: Union[float, list]=0) -> np.ndarray:
    # set offset to 2000
    offset = datetime.datetime(2000, 1, 1)
    td_array = np.array([])
    size = (
        np.size(year),
        np.size(month),
        np.size(day),
        np.size(UT),
        np.size(minutes)
    )
    max_size = max(size)
    if np.size(year) == 1:
        year = np.ones(max_size, dtype=int) * year
    if np.size(month) == 1:
        month = np.ones(max_size, dtype=int) * month
    if np.size(day) == 1:
        day = np.ones(max_size, dtype=int) * day
    if np.size(UT) == 1:
        UT = np.ones(max_size, dtype=int) * UT
    if np.size(minutes) == 1:
        minutes = np.ones(max_size, dtype=int) * minutes
    for i in range(np.size(year)):
        td = datetime.datetime(
            int(year[i]),
            int(month[i]),
            int(day[i]),
            int(UT[i]),
            int(minutes[i])
        ) - offset
        td_f = td.days+td.seconds/(3600.0*24)
        td_array = np.append(td_array, td_f)

    return td_array



import numpy as np
# from DIFI import get_SQ_field_TESTING_ONLY as getSQfield
from DIFI.getSQfield import getSQfield
import pytest

def parse_geomagnetic_data(filepath):
    # Define column labels
    keys = [
        "year", "month", "day", "hour", "min", "sec",
        "r", "theta", "phi", "h", "lat", "lon", "F10.7",
        "B_r", "B_theta", "B_phi", "X", "Y", "Z"
    ]

    data = []
    with open(filepath, 'r') as f:
        for line in f:
            # Split the line into values and convert to floats
            values = list(map(float, line.strip().split()))
            # Create a dictionary for each row
            entry = dict(zip(keys, values))
            data.append(entry)
    
    return data
def test_xdifi2_accuracy():
    with pytest.warns(UserWarning) as record:
        data = parse_geomagnetic_data("tests/test_values_xDIFI2_v2_aboveSQfield_20251015.txt")
        max_diff = 0.0

        for i in range(0, len(data)):
            B = getSQfield(data[i]['lat'], data[i]['lon'], data[i]['year'], data[i]['month'], data[i]['day'], hour=data[i]['hour'], minutes = data[i]['min'], h= data[i]['h'],f107_1 = data[i]['F10.7'], model_name='xdifi2')

            if B['X'] != data[i]['X']:
                max_diff = max(max_diff, np.abs(B['X'] - data[i]['X']))
            if B['Y'] != data[i]['Y']:
                max_diff = max(max_diff, np.abs(B['Y'] - data[i]['Y']))
            #     print("Y mismatch", B['Y'] , data[i]['Y'], B['Y'] - data[i]['Y'])
            if B['Z'] != data[i]['Z']:
                max_diff = max(max_diff, np.abs(B['Z'] - data[i]['Z']))

            if np.abs(B['X'] - data[i]['X']) > 1e-3:
                print("there is a mismatch in B['X'] - data[i]['X'] of size", B['X'] - data[i]['X'], B['X'] , data[i]['X'])
                raise ValueError("There was a mismatch of magnitude greater than 1e-3 in the X component")
            if np.abs(B['Y'] - data[i]['Y']) > 1e-3:
                print("there is a mismatch in B['Y'] - data[i]['Y'] of size", B['Y'] - data[i]['Y'], B['Y'] , data[i]['Y'])
                raise ValueError("There was a mismatch of magnitude greater than 1e-3 in the Y component")
            if np.abs(B['Z'] - data[i]['Z']) > 1e-3:
                print("there is a mismatch in B['Z'] - data[i]['Z'] of size", B['Z'] - data[i]['Z'], B['Z'] , data[i]['Z'])
                raise ValueError("There was a mismatch of magnitude greater than 1e-3 in the Y component")
        # print(f"Max difference: {max_diff}")
        # messages = [str(w.message) for w in record]
        # print("The thing inside messages:",messages)
        # expected_messages = [
        #     "Warning: The altitude is within 30 km of 110 km, where the model switches between internal and external ionospheric current sources. The model is not expected to accurately represent the field within this transition zone.', 'Warning: The altitude is within 30 km of 110 km, where the model switches between internal and external ionospheric current sources. The model is not expected to accurately represent the field within this transition zone.', 'Warning: The altitude is within 30 km of 110 km, where the model switches between internal and external ionospheric current sources. The model is not expected to accurately represent the field within this transition zone."
        # ]

        # assert messages == expected_messages[0]
        assert(max_diff < 1e-3)

# test_xdifi2_accuracy()


def SwarmL2_F107_Read(filename):
    # [t_f107,f107] = SwarmL2_F107_Read(filename)

    # Reads f107 values from DBL data file (Swarm L2 format).

    # Input:    filename        data filename

    # Output:   t_f107            time vector
    #           f107              data

    # A. Chulliat, 2016-09-19
    # (from earlier version: read_f107_new.m)

    t_f107 = []
    f107 = []
    fid = open(filename, 'r')

    for tline in fid:
        # skip comment lines
        if tline[0] == '#':
            print("comment line")
            continue
        # read data lines
        f = [float(i) for i in tline.split()]
        t_f107.append(f[0])
        f107.append(f[1])

    fid.close()

    return t_f107, f107



import numpy as np
from DIFI.getSQfield import getSQfield
import pytest

def parse_geomagnetic_data(filepath):
    # Define column labels
    keys = [
        "year", "month", "day", "hour", "min", "sec",
        "r", "theta", "phi", "h", "lat", "lon", "F10.7",
        "B_r", "B_theta", "B_phi", "X", "Y", "Z"
    ]

    data = []
    with open(filepath, 'r') as f:
        for line in f:
            # Split the line into values and convert to floats
            values = list(map(float, line.strip().split()))
            # Create a dictionary for each row
            entry = dict(zip(keys, values))
            data.append(entry)
    
    return data
def test_list_input_accuracy():
    models = ['xDIFI2', 'DIFI8']
    for model_name in models:
        data = parse_geomagnetic_data(f"tests/test_values_{model_name}_v1_20250928.txt")
        lat, lon, year, month, day, hour, minute, h, f107 = [],[],[],[],[],[],[],[],[]
        for i in range(0,len(data)):
            lat.append(data[i]['lat'])
            lon.append(data[i]['lon'])
            year.append(data[i]['year'])
            month.append(data[i]['month'])
            day.append(data[i]['day'])
            hour.append(data[i]['hour'])
            minute.append(data[i]['min'])
            h.append(data[i]['h'])
            f107.append(data[i]['F10.7'])


        B = getSQfield(lat, lon, year, 
                        month, day, hour=hour,
                            minutes = minute, h = h,f107_1 = f107, 
                            model_name = model_name)
        max_diff = 0.0
        for i in range(0, len(data)):
            # B = getSQfield(data[i]['lat'], data[i]['lon'], data[i]['year'], data[i]['month'], data[i]['day'], hour=data[i]['hour'], minutes = data[i]['min'], h= data[i]['h'],f107_1 = data[i]['F10.7'], model_name='xdifi2')

            if B['X'][i] != data[i]['X']:
                max_diff = max(max_diff, np.abs(B['X'][i] - data[i]['X']))
            if B['Y'][i] != data[i]['Y']:
                max_diff = max(max_diff, np.abs(B['Y'][i] - data[i]['Y']))
            #     print("Y mismatch", B['Y'][i] , data[i]['Y'], B['Y'][i] - data[i]['Y'])
            if B['Z'][i] != data[i]['Z']:
                max_diff = max(max_diff, np.abs(B['Z'][i] - data[i]['Z']))

            if np.abs(B['X'][i] - data[i]['X']) > 1e-3:
                print("there is a mismatch in B['X'][i] - data[i]['X'] of size", B['X'][i] - data[i]['X'])
                raise ValueError("There was a mismatch of magnitude greater than 1e-3 in the X component")
            if np.abs(B['Y'][i] - data[i]['Y']) > 1e-3:
                print("there is a mismatch in B['Y'][i] - data[i]['Y'] of size", B['Y'][i] - data[i]['Y'])
                raise ValueError("There was a mismatch of magnitude greater than 1e-3 in the Y component")
            if np.abs(B['Z'][i] - data[i]['Z']) > 1e-3:
                print("there is a mismatch in B['Z'][i] - data[i]['Z'][i] of size", B['Z'][i] - data[i]['Z'])
                raise ValueError("There was a mismatch of magnitude greater than 1e-3 in the Y component")
        print(f"Max difference of {model_name} list input: {max_diff}")
        assert(max_diff < 1e-3)



import numpy as np
from DIFI.getSQfield import getSQfield
import pytest

def parse_geomagnetic_data(filepath):
    # Define column labels
    keys = [
        "year", "month", "day", "hour", "min", "sec",
        "r", "theta", "phi", "h", "lat", "lon", "F10.7",
        "B_r", "B_theta", "B_phi", "X", "Y", "Z"
    ]

    data = []
    with open(filepath, 'r') as f:
        for line in f:
            # Split the line into values and convert to floats
            values = list(map(float, line.strip().split()))
            # Create a dictionary for each row
            entry = dict(zip(keys, values))
            data.append(entry)
    
    return data
def test_single_time_point_warnings():
    with pytest.warns(UserWarning) as record:
        #early warning: input date is just before the minimum recomended time
        B = getSQfield(0, 0, 2000, 
                        1, 1, h = 0,f107_1 = 100, 
                            model_name = 'xdifi2')

        #late warning: input date is just after the maximum recomended time
        B = getSQfield(0, 0, 2024, 
                        1, 2, h = 0,f107_1 = 100, 
                            model_name = 'xdifi2')
        #early warning: input date is just before the minimum recomended time
        B = getSQfield(0, 0, 2000, 
                        1, 1, h = 0,f107_1 = 100, 
                            model_name = 'difi8')
        #late warning: input date is just after the maximum recomended time
        B = getSQfield(0, 0, 2024, 
                        1, 2, h = 0,f107_1 = 100, 
                            model_name = 'difi8')
    assert len(record) == 4

    # Extract the actual warning messages
    messages = [str(w.message) for w in record]

    expected_messages = [
        "Dataset contains date before 2001.0, outside xDIFI2's recommended range",
        "Dataset contains date after 2024.0, outside xDIFI2's recommended range",
        "Dataset contains date before 2014.0, outside DIFI8's recommended range",
        "Dataset contains date after 2024.0, outside DIFI8's recommended range",
    ]

    assert messages == expected_messages


import os
import numpy as np
from functools import lru_cache
from DIFI import getSQfield, SwarmL2_F107_Read, SwarmL2_MIO_SHA_Read_v2
from typing import Optional, Union


@lru_cache(maxsize=1)
def load_coefs() -> tuple[list, list]:
    baseDir = os.path.dirname(__file__)
    filename_f107 = os.path.join(baseDir, "coefs", "f107.DBL")
    difi_t_f107, difi_f107 = SwarmL2_F107_Read.SwarmL2_F107_Read(filename_f107)

    return difi_t_f107, difi_f107
1
@lru_cache(maxsize=1)
def load_swarm_DIFI7() -> dict:
    """
    This function is called in this file to import 
    DIFI7 coefficients
    """
    baseDir = os.path.dirname(__file__)
    filename_DIFI = os.path.join(baseDir, "coefs", "difi-coefs.txt")
    swarm_data = SwarmL2_MIO_SHA_Read_v2.SwarmL2_MIO_SHA_Read_v2(filename_DIFI)

    return swarm_data

@lru_cache(maxsize=1)
def load_swarm_DIFI8() -> dict:
    baseDir = os.path.dirname(__file__)
    filename_DIFI = os.path.join(baseDir, "coefs", "DIFI8.DBL")
    swarm_data = SwarmL2_MIO_SHA_Read_v2.SwarmL2_MIO_SHA_Read_v2(filename_DIFI)

    return swarm_data

@lru_cache(maxsize=1)
def load_swarm_xDIFI() -> dict:
    baseDir = os.path.dirname(__file__)
    filename_DIFI = os.path.join(baseDir, "coefs", "xDIFI2.DBL")
    swarm_data = SwarmL2_MIO_SHA_Read_v2.SwarmL2_MIO_SHA_Read_v2(filename_DIFI)

    return swarm_data


# difi_t_f107, difi_f107 = load_coefs()
# swarm_data = load_swarm_DIFI7()

def get_f107_index(sq_t: list, start_time: float, end_time: float, difi_f107: Union[float,list], difi_t_f107: Union[float,list]) ->np.ndarray:

    frac_arr = (sq_t + .5) - np.floor(sq_t + .5)
    # frac_arr = (sq_t + 0) - np.floor(sq_t + 0)
    f107_1 = np.array([])

    for i in range(np.size(sq_t)):
        if sq_t[i] < start_time:
            raise Exception(
                "This package does not contain f10.7 data before 01/01/2000. Input time data contains a date corresponding to f10.7 data not contained in this package."
            )
        elif sq_t[i] > end_time:
            raise Exception(
                "This package does not contain f10.7 data after noon 12/31/2025. Input time data contains a date corresponding to f10.7 data not contained in this package."
            )
        while sq_t[i] < 0:
            sq_t[i] += 365
        j = 0
        while difi_t_f107[j] <= sq_t[i]:
            j += 1
        if sq_t[i] < .5:
            f107_1 = np.append(
            f107_1,
            difi_f107[j]
            )
        else:
            f107_1 = np.append(
                f107_1,
                difi_f107[j] * frac_arr[i] + difi_f107[j - 1] * (1 - frac_arr[i])
            )
            # f107_1 = np.append(
            #     f107_1,
            #     difi_f107[j] * (1-frac_arr[i]) + difi_f107[j - 1] * (frac_arr[i])
            # )

    return f107_1




# Partially autogenerated with SMOP 0.29
import numpy as np


def SwarmL2_MIO_SHA_Read_v2(filename: str) -> dict:
    # s = SwarmL2_MIO_SHA_Read(filename)

    # Purpose:
    # Read Swarm L2 MIO_SHA model from "*.DBL" file.

    # Input:
    #    filename           name of file to read

    # Output: s is a dictionary with the following elements
    #    m_e_d_Re(:,:)    primary Sq model (REAL coefficients)
    #    m_i_d_Re(:,:)    secondary Sq model (REAL coefficients)
    #    nmax, mmax       max degree and order in dipole coord.
    #    p_vec(:)         diurnal wavenumbers
    #    s_vec(:)         seasonal wavenumbers
    #    theta_NGP        co-latitude of North Geomagnetic dipole [deg]
    #    phi_NGP          longitude of North Geomagnetic dipole [deg]
    #    h                altitude of Sq current layer [km]
    #    N                Wold ratio of the F10.7 dependence

    # A. Chulliat, 2016-09-22
    # (from an earlier version dated 2011-03-27, with inputs from N. Olsen)
    # Python version rewritten by A. Woods, 2016-11-29
    s = {}
    # if exist(filename,r"file") == 0:

    try:
        fid = open(filename, 'r')
    except IOError as err:
        print("Coefficient file ", filename, r" does not exist!")
        raise err

    # skip comment lines

    tline = fid.readline()

    while tline[0] == r"#":
        tline = fid.readline()

    # read header part (1 line)
    # "%i %i %i %i %i %i %f %f %f %f"
    f = tline.split()
    s["nmax"] = int(f[0])
    s["mmax"] = int(f[1])

    pmin = int(f[2])
    pmax = int(f[3])
    smin = int(f[4])
    smax = int(f[5])

    s["theta_NGP"] = float(f[6])
    s["phi_NGP"] = float(f[7])
    s["h"] = float(f[8])
    s["N"] = float(f[9])
    s["p_vec"] = range(pmin, pmax + 1)
    s["s_vec"] = range(smin, smax + 1)

    N_nm = s["mmax"] * (s["mmax"] + 2) \
        + (s["nmax"] - s["mmax"]) * (2*s["mmax"] + 1)

    N_sp = len(s["p_vec"]) * len(s["s_vec"])

    # read primary model coefficients
    s["m_e_d_Re"] = np.zeros([N_nm, 2 * N_sp])
    for i in range(0, N_nm):
        tline = fid.readline()
        if tline.split() is None:
            raise ValueError(r"Wrong number of model coefficients!")
        f = [float(j) for j in tline.split()]
        s["m_e_d_Re"][i, :] = f[2:2*N_sp + 2]

    # read secondary model coefficients
    s["m_i_d_Re"] = np.zeros([N_nm, 2 * N_sp])
    for i in range(0, N_nm):
        tline = fid.readline()
        if tline.split() is []:
            raise ValueError(r"Wrong number of model coefficients!")
        f = [float(j) for j in tline.split()]

        s["m_i_d_Re"][i, :] = f[2:2*N_sp + 2]

    fid.close()

    return s



import math
import numpy as np
from typing import Union

# Celestial Computing with MATLAB but with Python
# input
# mjd2000 = modified julian day 2000
# output
# rasc = right ascension of the sun (radians)
#   (0 <= rasc <= 2 pi)
# decl = declination of the sun (radians)
#   (-pi/2 <= decl <= pi/2)
# rsun = eci position vector of the sun (km)
# note
#  coordinates are inertial, geocentric,
#  equatorial and true-of-date

# Modified by Nils Oslen, DSRI
# Translated to python by Adam Woods, NCEI


def sun_md2000(mjd2000: Union[float, list]) -> list[np.ndarray, np.ndarray]:
    atr = math.pi/648000
    # time arguments
    djd = mjd2000 - 0.5
    t = djd/36525.0 + 1

    # fundamental arguments (converted from revolutions to radians with r2r)
    gs = r2r(0.993126 + 0.0027377785 * djd)
    lm = r2r(0.606434 + 0.03660110129 * djd)
    ls = r2r(0.779072 + 0.00273790931 * djd)
    g2 = r2r(0.140023 + 0.00445036173 * djd)
    g4 = r2r(0.053856 + 0.00145561327 * djd)
    g5 = r2r(0.056531 + 0.00023080893 * djd)
    rm = r2r(0.347343 - 0.00014709391 * djd)

    # geocentric, ecliptic longitude of the sun (radians)
    plon = 6910 * np.sin(gs) + 72 * np.sin(2 * gs) - 17 * t * np.sin(gs)
    plon = plon - 7 * np.cos(gs - g5) + 6 * np.sin(lm - ls) \
        + 5 * np.sin(4 * gs - 8 * g4 + 3 * g5)
    plon = plon - 5 * np.cos(2 * (gs - g2)) - 4 * (
        np.sin(gs - g2)
        - np.cos(4 * gs - 8 * g4 + 3 * g5)
    )
    plon = plon + 3 * (
        np.sin(2 * (gs - g2)) - np.sin(g5) - np.sin(2 * (gs - g5))
    )
    plon = ls + atr * (plon - 17 * np.sin(rm))
    # plon%2pi*180/pi

    # geocentric distance of the sun (kilometers)
    # rsm = 149597870.691 * (
    #     1.00014 - 0.01675 * np.cos(gs) - 0.00014 * np.cos(2 * gs)
    # )

    # obliquity of the ecliptic (radians)
    obliq = atr * (84428 - 47 * t + 9 * np.cos(rm))

    # geocentric, equatorial right ascension and declination (radians)
    a = np.sin(plon)*np.cos(obliq)
    b = np.cos(plon)

    rasc = np.arctan2(a, b)
    decl = np.arcsin(np.sin(obliq)*np.sin(plon))

    return [rasc, decl]


def r2r(x: float) -> float:
    # revolutions to radians function
    # input
    #  x = argument(revolutions; 0 <= x <= 1)
    # returns
    #  equivalent x (radians; 0 <= y <= 2 pi)

    return 2 * math.pi * (x % 1)



import numpy as np
from DIFI.getSQfield import getSQfield
import pytest

def parse_geomagnetic_data(filepath):
    # Define column labels
    keys = [
        "year", "month", "day", "hour", "min", "sec",
        "r", "theta", "phi", "h", "lat", "lon", "F10.7",
        "B_r", "B_theta", "B_phi", "X", "Y", "Z"
    ]

    data = []
    with open(filepath, 'r') as f:
        for line in f:
            # Split the line into values and convert to floats
            values = list(map(float, line.strip().split()))
            # Create a dictionary for each row
            entry = dict(zip(keys, values))
            data.append(entry)
    
    return data

def test_proper_warnings_display():
    with pytest.warns(UserWarning) as record:
        #list input, double wanrning for values being above and below tolerance
        model_name = 'DIFI8'
        data = parse_geomagnetic_data(f"tests/test_values_{model_name}_v1_20250928.txt")
        lat, lon, year, month, day, hour, minute, h, f107 = [],[],[],[],[],[],[],[],[]
        for i in range(0,len(data)):
            lat.append(data[i]['lat'])
            lon.append(data[i]['lon'])
            year.append(data[i]['year'])
            month.append(data[i]['month'])
            day.append(data[i]['day'])
            hour.append(data[i]['hour'])
            minute.append(data[i]['min'])
            h.append(data[i]['h'])
            f107.append(data[i]['F10.7'])
        N = 40
        #Create time from 2000.0 to 2025.9
        lat, lon, year, month, day, hour, minute, h, f107 = lat[:N], lon[:N], year[:N], month[:N], day[:N], hour[:N], minute[:N], h[:N], f107[:N]
        years = np.linspace(2000, 2025, N)
        months = np.linspace(1,12, N)
        days = np.linspace(1,31, N)
        hours = np.linspace(0,23, N)
        minutes = np.linspace(0,59, N)

        B = getSQfield(lat, lon,years, 
                        months, days, hour= hours, minutes=minutes ,
                        h = 0, 
                        model_name = 'xdifi2')
        
        years = np.linspace(2014, 2024, N)
        B = getSQfield(lat, lon,years, 
                        months, days, hour= hours, minutes=minutes ,
                        h = 0,f107_1 = 100, 
                        model_name = 'difi8')
    assert len(record) == 3

    # Extract the actual warning messages
    messages = [str(w.message) for w in record]

    expected_messages = [
        "Dataset contains date before 2001.0, outside xDIFI2's recommended range",
        "Dataset contains date after 2024.0, outside xDIFI2's recommended range",
        "Dataset contains date after 2024.0, outside DIFI8's recommended range",
    ]

    for message,expected_message in zip(messages,expected_messages):
        assert message == expected_message


import unittest
from pathlib import Path
import csv  # from Python stdlib

import numpy as np

from DIFI.getSQfield import getSQfield
from DIFI import jd2000_dt
from DIFI import get_f107_index_all
import warnings
import pytest

# The data files will be in the same directory as this test script
inputs_file_path = Path(__file__).parent / "test_values_xDIFI2_v1_20250928.txt"



class test_getSQfield(unittest.TestCase):

    def setUp(self):
        #Read in the test points as a list of dictionaries (one dict per each row)
        self.csvfields = ['lat','lon','year','month','day','hour','x','y','z']
        inputs_file = inputs_file_path.resolve()
        
        self.points = self.parse_geomagnetic_data(inputs_file)

    def parse_csv_files(self, inputs_file):

        points = []
        with open(inputs_file, "r") as fin:
            for row in csv.DictReader(fin, fieldnames=self.csvfields):
                # Will read in as str, convert to float
                point = {key: float(val) for key, val in row.items()}
                # Add in inputs that aren't in CSV
                point['minutes'] = 0.
                point['h'] = 0.
                points.append(point)

        # Pick a (repeatable) random subset so it doesn't take forever
        rng = np.random.default_rng(42)
        subset_inds = rng.integers(low=0, high=len(points), size=60).tolist()
        points = [points[ind] for ind in subset_inds]

        return points
  
    def get_f107(self,point):
        """Helper function to get f107 by date (test data doesn't specify)"""
        difi_t_f107, difi_f107 = get_f107_index_all.load_coefs()
        start_time = 5114.0
        end_time = float(difi_t_f107[-1])
        sq_t = jd2000_dt.jd2000_dt(point['year'], 
                                   point['month'], 
                                   point['day'], 
                                   point['hour'],
                                    point['minutes'])
        f107 = get_f107_index_all.get_f107_index(sq_t, start_time, end_time, difi_f107, difi_t_f107)
        return f107.tolist()[0] #returns an array and we just want float

    def parse_geomagnetic_data(self, filepath):
        # Define column labels
        keys = [
            "year", "month", "day", "hour", "minutes", "sec",
            "r", "theta", "phi", "h", "lat", "lon", "F10.7",
            "B_r", "B_theta", "B_phi", "x", "y", "z"
        ]

        data = []
        with open(filepath, 'r') as f:
            for line in f:
                # Split the line into values and convert to floats
                values = list(map(float, line.strip().split()))
                # Create a dictionary for each row
                entry = dict(zip(keys, values))
                data.append(entry)

        return data

    def test_all_scalar(self):
        for point in self.points:
            Bxyz_test = getSQfield(lat=point['lat'],
                                lon=point['lon'],
                                year=point['year'],
                                month=point['month'],
                                day=point['day'],
                                hour=point['hour'],
                                minutes=point['minutes'],
                                h=point['h'],
                                f107_1=point["F10.7"])

            Bxyz_expected = {'X':point['x'],
                            'Y':point['y'],
                            'Z':point['z']}

            for cmpnt in ['X','Y','Z']:
                with self.subTest(f'B{cmpnt} {point}'):
                    B_test = Bxyz_test[cmpnt][0]
                    B_expect = Bxyz_expected[cmpnt]
                    self.assertAlmostEqual(B_test,B_expect,places=5)

    def test_all_lists(self):
        """Make sure we can handle all inputs as length 2 list"""
        point1,point2 = self.points[0],self.points[1]
        Bxyz_test = getSQfield(lat=[point1['lat'],point2['lat']],
                            lon=[point1['lon'],point2['lon']],
                            year=[point1['year'],point2['year']],
                            month=[point1['month'],point2['month']],
                            day=[point1['day'],point2['day']],
                            hour=[point1['hour'],point2['hour']],
                            minutes=[point1['minutes'],point2['minutes']],
                            h=[point1['h'],point2['h']],
                            f107_1=[point1['F10.7'],point2['F10.7']])

        Bxyz_expected = {'X':np.array([point1['x'],point2['x']]),
                        'Y':np.array([point1['y'],point2['y']]),
                        'Z':np.array([point1['z'],point2['z']])}

        for cmpnt in ['X','Y','Z']:
            with self.subTest(f'B{cmpnt} {point1} {point2}'):
                B_test = Bxyz_test[cmpnt].tolist()
                B_expect = Bxyz_expected[cmpnt].tolist()
                #point 1
                self.assertAlmostEqual(B_test[0],B_expect[0],places=6)
                #point 2
                self.assertAlmostEqual(B_test[1],B_expect[1],places=6)

    def test_inputs_scalar_except_one_list(self):
        """Check all possible permutations of scalars and one list
        (make sure all inputs can be a list when everything else is scalar)"""
        for i in range(len(self.points) // 3):
            point = self.points[i].copy() #make a copy so we don't overwrite original

            point['f107_1']=self.points[i]['F10.7']
            input_names = ['lat','lon','year','month','day','hour','minutes','h','f107_1']
            scalar_getSQfield_inputs = {key:point[key] for key in input_names}
            for input_name in input_names:
                with self.subTest(f'Test {input_name} is a list'):
                    getSQfield_inputs = scalar_getSQfield_inputs.copy()
                    
                    #Make one input a list
                    getSQfield_inputs[input_name] = [point[input_name],
                                                    point[input_name]]

                    Bxyz_test = getSQfield(**getSQfield_inputs)

                    #We expect to get 2 identical points back
                    Bxyz_expected = {'X':[point['x'],point['x']],
                                    'Y':[point['y'],point['y']],
                                    'Z':[point['z'],point['z']]}

                    for cmpnt in ['X','Y','Z']:
                        with self.subTest(f'B{cmpnt}'):
                            B_test = Bxyz_test[cmpnt]
                            B_expect = Bxyz_expected[cmpnt]
                            self.assertAlmostEqual(B_test[0],B_expect[0],places=5)
                            self.assertAlmostEqual(B_test[1],B_expect[1],places=5)





if __name__ == '__main__':
    unittest.main()




                            
        
            



import math
from typing import Union

import numpy as np

from DIFI import sun_md2000
def getmut(mjd2000: list, theta_NGP: Union[float, list], phi_NGP: float, *args, **kwargs) -> np.ndarray:
    # mut = getmut(mjd2000,theta_NGP,phi_NGP)

    # Calculate the Magnetic Universal Time for a given MJD2000 and a given
    # position of the North Geomagnetic Pole.

    # Input:    mjd2000         modified julian day 2000 (days)
    #           theta_NGP       geographic co-latitude of North Geomagnetic
    #                           Pole (deg)
    #           phi_NGP         geographic longitude of North Geomagnetic
    #                           Pole (deg)
    # Output:   mut             magnetic universal time (hr)

    # A. Chulliat, October 2008
    # Converted to python by A. Woods

    ######################################################################

    rad = math.pi / 180
    vec = np.ones(mjd2000.size)
    theta_0_rad = np.deg2rad(theta_NGP*vec)
    phi_0 = phi_NGP*vec
    ut = (mjd2000 - np.floor(mjd2000))*24
    phi_s = 15*(12*vec - ut)

    rasc, decl = sun_md2000.sun_md2000(mjd2000)
    theta_s = 90*vec - decl / rad

    theta_s_rad = np.deg2rad(theta_s)
    phi_diff_rad = np.deg2rad(phi_s - phi_0)

    eopp = np.sin(theta_s_rad)*np.sin(phi_diff_rad)
    eadj = np.cos(theta_0_rad)*np.sin(theta_s_rad)*np.cos(phi_diff_rad) \
        - np.sin(theta_0_rad)*np.cos(theta_s_rad)
    mut = 12*vec - np.arctan2(eopp, eadj) / rad / 15

    return mut



# Autogenerated with SMOP 0.29
import datetime, warnings
import numpy as np
from typing import Union

from DIFI import jd2000_dt
from DIFI import getmut
from DIFI import gg2gm_2010
from DIFI import design_SHA_Sq_i_Re_v2
from DIFI import design_SHA_Sq_e_Re_v2


def forward_Sq_d_Re(r: Union[float, list], theta: Union[float, list], phi: Union[float, list], t: Union[float, list], f107: Union[float, list], s: dict) -> tuple[np.ndarray, np.ndarray]:
    # [B_1,B_2] = forward_Sq_d_Re(r,theta,phi,t,f107,s)

    # Calculate the primary (1) and induced (2) Sq magnetic field from a model
    # in DIPOLE coordinates (parameterization of Sabaka et al., 2002).
    # The output values B_1 and B_2 are in SPHERICAL Coordinates

    # Inputs:
    #     r(:)                    radius [km]
    #     theta(:), phi(:)        geographic co-latitude, longitude [deg]
    #     t(:)                    time [MJD2000]
    #     f107(:)                 F10.7 index
    #     s is a structure with the following elements
    #         m_e_d_Re(:,:)           primary Sq model (REAL coefficients)
    #         m_i_d_Re(:,:)           secondary Sq model (REAL coefficients)
    #         nmax, mmax              max degree and order in dipole coord.
    #         p_vec(:)                diurnal wavenumbers
    #         s_vec(:)                seasonal wavenumbers
    #         theta_NGP               co-latitude of N. Geomagnetic Pole [deg]
    #         phi_NGP                 longitude of N. Geomagnetic Pole [deg]
    #         h                       altitude of Sq currents [km]
    #         N                       Woldf ratio of the F10.7 dependence

    # Dependendies :
    # - "gg2gm", "jd2000", "getmut"
    # - "design_SHA_Sq_i_Re" and "design_SHA_Sq_e_Re"

    # A. Chulliat, 2016-09-22
    # (from an earlier version dated 2011-04-23, with inputs from N. Olsen)

    # Translated to Python by A. Woods 2018-1-1

    # Switchable 20km altitude validity limit added by L. Kilcommons 2025-04-07

    # calculate radii in units of reference radius
    a = 6371.2
    rho = r / a
    rho_20km = (a + 20.) / a
    rho_Sq = (a + s['h']) / a
    # convert to matrix if input parameter is scalar
    max_size = max(
        np.array(
            (
                [np.size(rho)],
                [np.size(theta)],
                [np.size(phi)],
                [np.size(t)],
                [np.size(f107)]
            )
        )
    )
    if np.isscalar(rho) or len(rho) == 1:
        rho = np.multiply(rho, np.ones(max_size))

    if np.isscalar(theta) or len(theta) == 1:
        theta = np.multiply(theta, np.ones(max_size))

    if np.isscalar(phi) or len(phi) == 1:
        phi = np.multiply(phi, np.ones(max_size))

    if np.isscalar(t) or len(t) == 1:
        t = np.multiply(t, np.ones(max_size))

    if np.isscalar(f107) or len(f107) == 1:
        f107 = np.multiply(f107, np.ones(max_size))

    if (
        np.size(t) != np.size(rho)
        or np.size(t) != np.size(theta)
        or np.size(t) != np.size(phi)
        or np.size(t) != np.size(f107)
    ):
        print(
            np.size(t),
            np.size(rho),
            np.size(theta),
            np.size(phi),
            np.size(f107),
        )
        raise Exception(r"Variables must be of equal size (or scalars)")

    w_s = 2*np.pi
    w_p = 2*np.pi / 24
    N_data = np.size(theta, 0)

    # calculate time in year (season) and MUT
    t_1 = [
        datetime.timedelta(int(t_iter), (t_iter - int(t_iter)) * 3600)
        + datetime.datetime(2000, 1, 1)
        for t_iter in t
    ]

    year = [t_iter.year for t_iter in t_1]
    ndays = np.zeros(np.size(year))
    for i in range(len(year)):
        ndays[i] = (
            datetime.datetime(year[i]+1, 1, 1)
            - datetime.datetime(year[i], 1, 1)
        ).days
    # list comprehension must be converted to ndarray
    t_season = np.array(
        [
            (t[i] - jd2000_dt.jd2000_dt(year[i], 1, 1, 0)) / ndays[i]
            for i in range(len(ndays))
        ]
    )
    # not inlcuded in HDGM version
    t_season = t_season.flatten()
    t_mut = getmut.getmut(t, s['theta_NGP'], s['phi_NGP'])
    # calculate dipolar coordinates + matrix R
    theta_d, phi_d, rotmat = gg2gm_2010.gg2gm_2010(theta, phi, get_R=True, theta_b=s['theta_NGP'], phi_b = s['phi_NGP'])

    arr_internal = np.array(
        design_SHA_Sq_i_Re_v2.design_SHA_Sq_i_Re_v2(
            rho,
            theta_d,
            phi_d,
            s['nmax'],
            s['mmax'],
        )
    )

    s_vec = np.array(s['s_vec'])
    p_vec = np.array(s['p_vec'])

    beta = (
        w_s * s_vec[:, None, None] * t_season[None, None, :]
        + w_p * p_vec[None, :, None] * t_mut[None, None, :]
    )
    beta = beta.reshape(
        -1, N_data
    )
    time_arr = np.array(
        [
            np.cos(beta),
            np.sin(beta),
        ]
    )
    time_arr = time_arr.transpose(1, 0, 2).reshape(-1, N_data)

    # CASE #1: above Sq currents
    if (min(rho) > rho_Sq):
        #=========Above SQ coeff shift=======
        mmax = s['mmax']
        nmax = s['nmax']
        N_nm = mmax * (mmax + 2) + (nmax - mmax) * (2 * mmax + 1)
        f = np.zeros((N_nm))  
        i1 = 0  

        for n in range(1, nmax + 1):
            nr = min(2 * n + 1, 2 * mmax + 1)
            f[i1:i1 + nr] = -n / (n + 1) * rho_Sq ** (2 * n + 1)
            i1 += nr
        N_sp = len(p_vec) * len(s_vec)
        f = np.tile(f, (2 * N_sp, 1))
        N_coeff_nm = 2 * N_nm * N_sp
        coeff_shape = np.shape(s['m_e_d_Re'])
        f = f.flatten()
        m_e_d_Re = np.reshape(s['m_e_d_Re'],N_coeff_nm,'F')
        m_e_d_Re_C = m_e_d_Re * f  # elementwise multiply (broadcasts like bsxfun)
        
        m_e_d_Re_C = np.reshape(m_e_d_Re_C,coeff_shape,'F')

        #=========Above SQ coeff shift end=======
        B_1_tmp = np.einsum(
            'ij, lik, jk ->lk',
            m_e_d_Re_C,#s['m_e_d_Re'],
            arr_internal,
            time_arr,
        )

        B_2_tmp = np.einsum(
            'ij, lik, jk ->lk',
            s['m_i_d_Re'],
            arr_internal,
            time_arr,
        )
        
    # CASE #2: below Sq currents
    elif (max(rho) < rho_Sq):
        arr_external = np.array(
            design_SHA_Sq_e_Re_v2.design_SHA_Sq_e_Re_v2(
                rho,
                theta_d,
                phi_d,
                s['nmax'],
                s['mmax'],
            )
        )

        B_1_tmp = np.einsum(
            'ij, lik, jk ->lk',
            s['m_e_d_Re'],
            arr_external,
            time_arr,
        )

        B_2_tmp = np.einsum(
            'ij, lik, jk ->lk',
            s['m_i_d_Re'],
            arr_internal,
            time_arr,
        )

    # CASE #3: error
    else:
        raise Exception(r"data in both regions (below and above Sq currents)")

    # Rotate into geomagnetic frame
    # fix minus sign
    B_1_tmp[[1, 2]] = np.einsum(
        'kij, jk -> ik',
        rotmat.transpose(0, 2, 1),
        B_1_tmp[[1, 2]],
    )

    B_2_tmp[[1, 2]] = np.einsum(
        'kij, jk -> ik',
        rotmat.transpose(0, 2, 1),
        B_2_tmp[[1, 2]],
    )

    # correct for F10.7 dependence
    w = (1 + s['N']*f107)
    B_1 = B_1_tmp * w
    B_2 = B_2_tmp * w
    return B_1, B_2 
